/*
   Billy the Kid (btk_obj.c)
   "..a python module to do all kinds of raw network shit..."

   by Gorny (gorny0@zonnet.nl)
   http://gorny.cjb.net
*/

#include "wrap.h"
#include "packet.h"
#include "btk_obj.h"

btk_object *
new_btk_object(PyObject *args)
{
        btk_object *self;
        self = PyObject_New(btk_object, &btk_object_type);
        if (self == NULL)
                return NULL;
        self->btk_obj_attr = NULL;
        return self;
}

static void
btk_object_dealloc(btk_object *self)
{
        Py_XDECREF(self->btk_obj_attr);
        PyObject_Del(self);
}

static PyObject *
btk_object_getattr(btk_object *self, char *name)
{
        if (self->btk_obj_attr != NULL) {
                PyObject *v = PyDict_GetItemString(self->btk_obj_attr, name);
                if (v != NULL) {
                        Py_INCREF(v);
                        return v;
                }
        }
        return Py_FindMethod(btk_obj_methods, (PyObject *)self, name);
}

static int
btk_object_setattr(btk_object *self, char *name, PyObject *v)
{
        if (self->btk_obj_attr == NULL) {
                self->btk_obj_attr = PyDict_New();
                if (self->btk_obj_attr == NULL)
                        return -1;
        }
        if (v == NULL) {
                int rv = PyDict_DelItemString(self->btk_obj_attr, name);
                if (rv < 0)
                        PyErr_SetString(PyExc_AttributeError,
                                "delete non-existing btk_obj attribute");
                return rv;
        }
        else
                return PyDict_SetItemString(self->btk_obj_attr, name, v);
}

/* setting references to automatic methods */
statichere PyTypeObject btk_object_type = {
        PyObject_HEAD_INIT(NULL)
        0,                      /*ob_size*/
        "btk_object",
        sizeof(btk_object),
        0,                      /*tp_itemsize*/
        (destructor)btk_object_dealloc,
        0,                      /*tp_print*/
        (getattrfunc)btk_object_getattr,
        (setattrfunc)btk_object_setattr,
        0,                      /*tp_compare*/
        0,                      /*tp_repr*/
        0,                      /*tp_as_number*/
        0,                      /*tp_as_sequence*/
        0,                      /*tp_as_mapping*/
        0,                      /*tp_hash*/
};

/*  
*  from here all btk_object methods are defined
*/
static PyObject * 
btk_object_protocol(btk_object *self, PyObject *args)
{
        int protocol;
        if (!PyArg_ParseTuple(args, "i:protocol", &protocol))
                return NULL;
	if (protocol == TCP) {
		self->protocol = TCP;
	        self->iph=(struct ip *) self->packet;
       	 	self->pseudo=(struct pseudo *)(self->packet + 
			   	sizeof(struct ip));
        	self->tcph=(struct tcphdr *) (self->packet + sizeof(struct ip)
                                + sizeof(struct pseudo));
        	self->data=(char *) (self->packet + sizeof(struct ip) +
                     		sizeof(struct pseudo) + sizeof(struct tcphdr));
		memset(self->data, '\0', strlen(self->data));
	}
}

static PyObject *
btk_object_data(btk_object *self, PyObject *args)
{
	char * data;
	if (!PyArg_ParseTuple(args, "s:data", &data))
		return NULL;
	self->data = data;
	if (self->data != data)
		return PyInt_FromLong(-1);
	return PyString_FromString(self->data);	
}

static PyObject *
btk_object_flags(btk_object *self, PyObject *args)
{
	int flags;
	if (!PyArg_ParseTuple(args, "i:flags", &flags))
		return NULL;
	self->flags = flags;
	if (self->flags != flags || self->flags < 0)
		return PyInt_FromLong(-1);
	Py_INCREF(Py_None);
	return Py_None;	
}

static PyObject *
btk_object_send(btk_object *self, PyObject *args)
{
	struct sockaddr_in foo;
        int rawfd, one;
	u_long length = strlen(self->data);

        if (!PyArg_ParseTuple(args, "sisi:send", &self->dstip, &self->dstport,
					&self->srcip, &self->srcport))
                return NULL;

	/* erasing packet */
        memset(self->packet,'\0', sizeof(self->packet));

	switch(self->protocol) {
	case TCP:
		fill_tcp_headers(self);
		break;
	case UDP:
		fill_udp_headers(self);
		break;
	case ICMP:
		fill_icmp_headers(self);
		break;
	};

	memset(&foo,'\0',sizeof(foo));
	foo.sin_family=AF_INET;
        foo.sin_addr.s_addr=inet_addr(self->dstip);
        one = 1;

        if ((rawfd=socket(PF_INET,SOCK_RAW,IPPROTO_ICMP))<0) {
                perror("RawSocket");
        }
        if (setsockopt(rawfd,IPPROTO_IP,IP_HDRINCL,&one,sizeof(one))<0) {
                perror("SetSockOpt");
        }

        if (sendto(rawfd,self->packet,sizeof(struct ip) +
                        sizeof(struct tcphdr)+length, 0,
                        (struct sockaddr *) &foo,sizeof(foo))<0)
		perror("SendingSock");

        Py_INCREF(Py_None);
        return Py_None;
}

static PyObject *
btk_object_options(btk_object *self, PyObject *args, PyObject *keywords)
{
/* making options not directly go to self->tcph / self->iph 'cause of problems
* can arise if btk_object_protocol() hasn't been run yet. */

	static char *kwlist[] = {"seq", "ack", "urp", "win", "off", "tos",
				 "ttl", "id", NULL};

	if (!PyArg_ParseTupleAndKeywords(args, keywords, "|iiiiiiii", kwlist,
			&self->seq, &self->ack, &self->urp, &self->win,
			&self->off, &self->tos, &self->ttl, &self->id))
        	return NULL;

	Py_INCREF(Py_None);
	return Py_None;
}

static PyObject *
btk_object_test(btk_object *self, PyObject *args)
{
        struct sockaddr_in foo;
        struct in_addr srcip, dstip;
        int rawfd, one;
        u_long length = strlen(self->data);

        memset(self->packet,'\0', sizeof(self->packet));
        self->pseudo->protocol=self->protocol;
        self->pseudo->length = htons(sizeof (struct tcphdr)+ length);
        self->iph->ip_hl = 5;        /* header length */
        self->iph->ip_v = 4;         /* version */
        self->iph->ip_id = 1234;      /* identification */
        self->iph->ip_len = sizeof(struct ip) + sizeof(struct tcphdr) + length;
        self->iph->ip_off = 0;
        self->iph->ip_tos = 0; /* not set while most routers don't bother */
        self->iph->ip_ttl = 40;
        self->iph->ip_p = (u_int8_t)self->protocol;
        self->tcph->th_x2 = 0;
        self->tcph->th_off = 5;
        self->tcph->th_flags = self->flags;
        self->tcph->th_win = 65535;
        self->tcph->th_urp = 0;
        self->tcph->th_ack = 0;
        self->tcph->th_seq = random();

        srcip.s_addr = inet_addr("127.0.0.1");
        dstip.s_addr = inet_addr("127.0.0.1");

        bzero(&foo,sizeof(foo));

        self->pseudo->saddr = srcip.s_addr;
        self->pseudo->daddr = dstip.s_addr;
        self->iph->ip_src = srcip;
        self->iph->ip_dst = dstip;
        self->tcph->th_sport = htons(1234);
        self->tcph->th_dport = htons(4321);

	if (self->data && length)
		memcpy(self->packet + sizeof(struct ip) + sizeof(struct tcphdr)+sizeof(struct pseudo), self->data, length);

        self->tcph->th_sum = in_cksum((unsigned short *) self->pseudo,
                                sizeof(struct pseudo)+length+sizeof(struct tcphdr));

	memcpy(self->pseudo, self->tcph,sizeof(struct tcphdr)+length);
        //bcopy(self->tcph,self->pseudo,sizeof(struct tcphdr)+length);
	//bcopy(self->data, self->packet+sizeof(struct iphdr)+sizeof(struct tcphdr), length);
	
	foo.sin_family=AF_INET;
        foo.sin_addr.s_addr=dstip.s_addr;

        one = 1;
        if ((rawfd=socket(PF_INET,SOCK_RAW,IPPROTO_ICMP))<0) {
                perror("RawSocket");
        }
        if (setsockopt(rawfd,IPPROTO_IP,IP_HDRINCL,&one,sizeof(one))<0) {
                perror("SetSockOpt");
        }
        if (sendto(rawfd,self->packet,sizeof(struct ip) +
                        sizeof(struct tcphdr)+length, 0,
                        (struct sockaddr *) &foo,sizeof(foo))<0)
		perror("wtf\n");

        Py_INCREF(Py_None);
        return Py_None;
}
